/*
 * Copyright (c) 2014-2015, TAKAHASHI Tomohiro (TTRFTECH) edy555@gmail.com
 * All rights reserved.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3, or (at your option)
 * any later version.
 *
 * The software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifdef __USE_CMSIS
#include "LPC43xx.h"
#endif

#include <cr_section_macros.h>
#include <limits.h>
#include <arm_math.h>
#include "lpc43xx_i2s.h"
#include "receiver.h"

int16_t finenco_phasestep;
uint16_t finenco_phase;

__RAMFUNC(RAM)
void nco_set_frequency(int32_t freq, float32_t ampl)
{
	int16_t *costbl = NCO_SIN_TABLE;
	int16_t *sintbl = NCO_COS_TABLE;
	int32_t f;
	int i;

	freq -= (int)(freq / ADC_RATE) * ADC_RATE;
	f = ((int64_t)freq * NCO_TABLE_SIZE + ADC_RATE/2) / ADC_RATE;
	for (i = 0; i < NCO_SAMPLES; i++) {
		float32_t phase = 2*PI*f*i/NCO_TABLE_SIZE;
		costbl[i] = (int16_t)(arm_cos_f32(phase) * ampl);
		sintbl[i] = (int16_t)(arm_sin_f32(phase) * ampl);
	}

	{
		int32_t tune = freq - (f * ADC_RATE)/NCO_TABLE_SIZE;
	    //printf("tune: %d\n", tune);
		finenco_phasestep = tune * 65536 / IF_RATE;
	    //printf("step: %d\n", finenco_phasestep);
	}
}


cic_state_t cic_i;
cic_state_t cic_q;

static void cic_init(void)
{
	memset(&cic_i, 0, sizeof cic_i);
	memset(&cic_q, 0, sizeof cic_q);
	cic_i.dest = I_FIR_BUFFER;
	cic_i.nco_base = NCO_SIN_TABLE;
	cic_q.dest = Q_FIR_BUFFER;
	cic_q.nco_base = NCO_COS_TABLE;
}

int32_t fir_gain = 4;
int32_t fir_rshift = 0;

__RAMFUNC(RAM)
void cic_decimate(cic_state_t *cic, uint8_t *buf, int len)
{
	int16_t *const dest = (int16_t*)cic->dest;
	uint32_t *const capture = (uint32_t*)buf;
	const uint32_t *const nco_base = (uint32_t*)cic->nco_base;

	int32_t s0 = cic->s0;
	int32_t s1 = cic->s1;
	int32_t d0 = cic->d0;
	int32_t d1 = cic->d1;
	int32_t e0, e1;
	int i, l;

	l = cic->dest_idx;
	for (i = 0; i < len / 4; ) {
		int j;
		for (j = 0; j < NCO_SAMPLES/2; ) {
#define CIC0()	do { \
			uint32_t f = nco_base[j++]; \
			uint32_t x = capture[i++]; \
			s0 = __SMLAD(x, f, s0); \
			s1 += s0; \
		} while(0)
#define CIC5()	CIC0();CIC0();CIC0();CIC0();CIC0()

			CIC5();CIC5();CIC5();CIC5();CIC5();
			CIC5();CIC5();CIC5();CIC5();CIC5();
			CIC5();CIC5();CIC5();CIC5();CIC5();

			e0 = d0 - s1;
			d0 = s1;
			e1 = d1 - e0;
			d1 = e0;
			e1 >>= fir_rshift;
			dest[l++] = __SSAT((e1 * fir_gain) >> 16, 16);
			l %=  FIR_BUFFER_SIZE/2;
		}
	}
	cic->dest_idx = l;
	cic->s0 = s0;
	cic->s1 = s1;
	cic->d0 = d0;
	cic->d1 = d1;
}



const int16_t cos_sin_table[256][4] = {
    { -32767,   10,      0, -804 },
    { -32757,   29,   -804, -804 },
    { -32728,   50,  -1608, -802 },
    { -32678,   69,  -2410, -802 },
    { -32609,   88,  -3212, -799 },
    { -32521,  109,  -4011, -797 },
    { -32412,  127,  -4808, -794 },
    { -32285,  148,  -5602, -791 },
    { -32137,  166,  -6393, -786 },
    { -31971,  186,  -7179, -783 },
    { -31785,  205,  -7962, -777 },
    { -31580,  224,  -8739, -773 },
    { -31356,  243,  -9512, -766 },
    { -31113,  261, -10278, -761 },
    { -30852,  281, -11039, -754 },
    { -30571,  298, -11793, -746 },
    { -30273,  317, -12539, -740 },
    { -29956,  335, -13279, -731 },
    { -29621,  353, -14010, -722 },
    { -29268,  370, -14732, -714 },
    { -28898,  388, -15446, -705 },
    { -28510,  405, -16151, -695 },
    { -28105,  422, -16846, -684 },
    { -27683,  438, -17530, -674 },
    { -27245,  455, -18204, -664 },
    { -26790,  471, -18868, -651 },
    { -26319,  487, -19519, -640 },
    { -25832,  503, -20159, -628 },
    { -25329,  518, -20787, -616 },
    { -24811,  532, -21403, -602 },
    { -24279,  548, -22005, -589 },
    { -23731,  561, -22594, -576 },
    { -23170,  576, -23170, -561 },
    { -22594,  589, -23731, -548 },
    { -22005,  602, -24279, -532 },
    { -21403,  616, -24811, -518 },
    { -20787,  628, -25329, -503 },
    { -20159,  640, -25832, -487 },
    { -19519,  651, -26319, -471 },
    { -18868,  664, -26790, -455 },
    { -18204,  674, -27245, -438 },
    { -17530,  684, -27683, -422 },
    { -16846,  695, -28105, -405 },
    { -16151,  705, -28510, -388 },
    { -15446,  714, -28898, -370 },
    { -14732,  722, -29268, -353 },
    { -14010,  731, -29621, -335 },
    { -13279,  740, -29956, -317 },
    { -12539,  746, -30273, -298 },
    { -11793,  754, -30571, -281 },
    { -11039,  761, -30852, -261 },
    { -10278,  766, -31113, -243 },
    { -9512,  773, -31356, -224 },
    { -8739,  777, -31580, -205 },
    { -7962,  783, -31785, -186 },
    { -7179,  786, -31971, -166 },
    { -6393,  791, -32137, -148 },
    { -5602,  794, -32285, -127 },
    { -4808,  797, -32412, -109 },
    { -4011,  799, -32521,  -88 },
    { -3212,  802, -32609,  -69 },
    { -2410,  802, -32678,  -50 },
    { -1608,  804, -32728,  -29 },
    {  -804,  804, -32757,  -10 },
    {     0,  804, -32767,   10 },
    {   804,  804, -32757,   29 },
    {  1608,  802, -32728,   50 },
    {  2410,  802, -32678,   69 },
    {  3212,  799, -32609,   88 },
    {  4011,  797, -32521,  109 },
    {  4808,  794, -32412,  127 },
    {  5602,  791, -32285,  148 },
    {  6393,  786, -32137,  166 },
    {  7179,  783, -31971,  186 },
    {  7962,  777, -31785,  205 },
    {  8739,  773, -31580,  224 },
    {  9512,  766, -31356,  243 },
    { 10278,  761, -31113,  261 },
    { 11039,  754, -30852,  281 },
    { 11793,  746, -30571,  298 },
    { 12539,  740, -30273,  317 },
    { 13279,  731, -29956,  335 },
    { 14010,  722, -29621,  353 },
    { 14732,  714, -29268,  370 },
    { 15446,  705, -28898,  388 },
    { 16151,  695, -28510,  405 },
    { 16846,  684, -28105,  422 },
    { 17530,  674, -27683,  438 },
    { 18204,  664, -27245,  455 },
    { 18868,  651, -26790,  471 },
    { 19519,  640, -26319,  487 },
    { 20159,  628, -25832,  503 },
    { 20787,  616, -25329,  518 },
    { 21403,  602, -24811,  532 },
    { 22005,  589, -24279,  548 },
    { 22594,  576, -23731,  561 },
    { 23170,  561, -23170,  576 },
    { 23731,  548, -22594,  589 },
    { 24279,  532, -22005,  602 },
    { 24811,  518, -21403,  616 },
    { 25329,  503, -20787,  628 },
    { 25832,  487, -20159,  640 },
    { 26319,  471, -19519,  651 },
    { 26790,  455, -18868,  664 },
    { 27245,  438, -18204,  674 },
    { 27683,  422, -17530,  684 },
    { 28105,  405, -16846,  695 },
    { 28510,  388, -16151,  705 },
    { 28898,  370, -15446,  714 },
    { 29268,  353, -14732,  722 },
    { 29621,  335, -14010,  731 },
    { 29956,  317, -13279,  740 },
    { 30273,  298, -12539,  746 },
    { 30571,  281, -11793,  754 },
    { 30852,  261, -11039,  761 },
    { 31113,  243, -10278,  766 },
    { 31356,  224,  -9512,  773 },
    { 31580,  205,  -8739,  777 },
    { 31785,  186,  -7962,  783 },
    { 31971,  166,  -7179,  786 },
    { 32137,  148,  -6393,  791 },
    { 32285,  127,  -5602,  794 },
    { 32412,  109,  -4808,  797 },
    { 32521,   88,  -4011,  799 },
    { 32609,   69,  -3212,  802 },
    { 32678,   50,  -2410,  802 },
    { 32728,   29,  -1608,  804 },
    { 32757,   10,   -804,  804 },
    { 32767,  -10,      0,  804 },
    { 32757,  -29,    804,  804 },
    { 32728,  -50,   1608,  802 },
    { 32678,  -69,   2410,  802 },
    { 32609,  -88,   3212,  799 },
    { 32521, -109,   4011,  797 },
    { 32412, -127,   4808,  794 },
    { 32285, -148,   5602,  791 },
    { 32137, -166,   6393,  786 },
    { 31971, -186,   7179,  783 },
    { 31785, -205,   7962,  777 },
    { 31580, -224,   8739,  773 },
    { 31356, -243,   9512,  766 },
    { 31113, -261,  10278,  761 },
    { 30852, -281,  11039,  754 },
    { 30571, -298,  11793,  746 },
    { 30273, -317,  12539,  740 },
    { 29956, -335,  13279,  731 },
    { 29621, -353,  14010,  722 },
    { 29268, -370,  14732,  714 },
    { 28898, -388,  15446,  705 },
    { 28510, -405,  16151,  695 },
    { 28105, -422,  16846,  684 },
    { 27683, -438,  17530,  674 },
    { 27245, -455,  18204,  664 },
    { 26790, -471,  18868,  651 },
    { 26319, -487,  19519,  640 },
    { 25832, -503,  20159,  628 },
    { 25329, -518,  20787,  616 },
    { 24811, -532,  21403,  602 },
    { 24279, -548,  22005,  589 },
    { 23731, -561,  22594,  576 },
    { 23170, -576,  23170,  561 },
    { 22594, -589,  23731,  548 },
    { 22005, -602,  24279,  532 },
    { 21403, -616,  24811,  518 },
    { 20787, -628,  25329,  503 },
    { 20159, -640,  25832,  487 },
    { 19519, -651,  26319,  471 },
    { 18868, -664,  26790,  455 },
    { 18204, -674,  27245,  438 },
    { 17530, -684,  27683,  422 },
    { 16846, -695,  28105,  405 },
    { 16151, -705,  28510,  388 },
    { 15446, -714,  28898,  370 },
    { 14732, -722,  29268,  353 },
    { 14010, -731,  29621,  335 },
    { 13279, -740,  29956,  317 },
    { 12539, -746,  30273,  298 },
    { 11793, -754,  30571,  281 },
    { 11039, -761,  30852,  261 },
    { 10278, -766,  31113,  243 },
    {  9512, -773,  31356,  224 },
    {  8739, -777,  31580,  205 },
    {  7962, -783,  31785,  186 },
    {  7179, -786,  31971,  166 },
    {  6393, -791,  32137,  148 },
    {  5602, -794,  32285,  127 },
    {  4808, -797,  32412,  109 },
    {  4011, -799,  32521,   88 },
    {  3212, -802,  32609,   69 },
    {  2410, -802,  32678,   50 },
    {  1608, -804,  32728,   29 },
    {   804, -804,  32757,   10 },
    {     0, -804,  32767,  -10 },
    {  -804, -804,  32757,  -29 },
    { -1608, -802,  32728,  -50 },
    { -2410, -802,  32678,  -69 },
    { -3212, -799,  32609,  -88 },
    { -4011, -797,  32521, -109 },
    { -4808, -794,  32412, -127 },
    { -5602, -791,  32285, -148 },
    { -6393, -786,  32137, -166 },
    { -7179, -783,  31971, -186 },
    { -7962, -777,  31785, -205 },
    { -8739, -773,  31580, -224 },
    { -9512, -766,  31356, -243 },
    { -10278, -761,  31113, -261 },
    { -11039, -754,  30852, -281 },
    { -11793, -746,  30571, -298 },
    { -12539, -740,  30273, -317 },
    { -13279, -731,  29956, -335 },
    { -14010, -722,  29621, -353 },
    { -14732, -714,  29268, -370 },
    { -15446, -705,  28898, -388 },
    { -16151, -695,  28510, -405 },
    { -16846, -684,  28105, -422 },
    { -17530, -674,  27683, -438 },
    { -18204, -664,  27245, -455 },
    { -18868, -651,  26790, -471 },
    { -19519, -640,  26319, -487 },
    { -20159, -628,  25832, -503 },
    { -20787, -616,  25329, -518 },
    { -21403, -602,  24811, -532 },
    { -22005, -589,  24279, -548 },
    { -22594, -576,  23731, -561 },
    { -23170, -561,  23170, -576 },
    { -23731, -548,  22594, -589 },
    { -24279, -532,  22005, -602 },
    { -24811, -518,  21403, -616 },
    { -25329, -503,  20787, -628 },
    { -25832, -487,  20159, -640 },
    { -26319, -471,  19519, -651 },
    { -26790, -455,  18868, -664 },
    { -27245, -438,  18204, -674 },
    { -27683, -422,  17530, -684 },
    { -28105, -405,  16846, -695 },
    { -28510, -388,  16151, -705 },
    { -28898, -370,  15446, -714 },
    { -29268, -353,  14732, -722 },
    { -29621, -335,  14010, -731 },
    { -29956, -317,  13279, -740 },
    { -30273, -298,  12539, -746 },
    { -30571, -281,  11793, -754 },
    { -30852, -261,  11039, -761 },
    { -31113, -243,  10278, -766 },
    { -31356, -224,   9512, -773 },
    { -31580, -205,   8739, -777 },
    { -31785, -186,   7962, -783 },
    { -31971, -166,   7179, -786 },
    { -32137, -148,   6393, -791 },
    { -32285, -127,   5602, -794 },
    { -32412, -109,   4808, -797 },
    { -32521,  -88,   4011, -799 },
    { -32609,  -69,   3212, -802 },
    { -32678,  -50,   2410, -802 },
    { -32728,  -29,   1608, -804 },
    { -32757,  -10,    804, -804 }
};

static inline
uint32_t cos_sin(uint16_t phase, int shift)
{
    uint16_t idx = phase / 256;
    uint16_t mod = phase & 0xff;
    uint32_t r = __PKHBT(0x0100, mod, 16);
    uint32_t *e = (uint32_t *)&cos_sin_table[idx];
    uint32_t cd = e[0];
    uint32_t sd = e[1];
    int32_t c = __SMUAD(r, cd);
    int32_t s = __SMUAD(r, sd);
    c /= 256;
    s /= 256;
    return __PKHBT(s, c, 16);
}

__RAMFUNC(RAM)
void finetune_mixer()
{
	uint32_t *src1 = (uint32_t *)I_FIR_BUFFER;
	uint32_t *src2 = (uint32_t *)Q_FIR_BUFFER;
	int size = FIR_BUFFER_SIZE;
	while (size > 0) {
		uint32_t cossin0 = cos_sin(finenco_phase, 0);
		finenco_phase += finenco_phasestep;
		uint32_t cossin1 = cos_sin(finenco_phase, 0);
		finenco_phase += finenco_phasestep;
		uint32_t i0i1 = *src1;
		uint32_t q0q1 = *src2;
		uint32_t i0q0 = __PKHTB(i0i1, q0q1, 16);
		uint32_t i1q1 = __PKHBT(q0q1, i0i1, 16);
		int32_t i0 = __SMLSD(i0q0, cossin0, 0);
		int32_t i1 = __SMLSD(i1q1, cossin1, 0);
		*src1++ = __PKHTB(i0, i1, 16);
		int32_t q0 = __SMLADX(i0q0, cossin0, 0);
		int32_t q1 = __SMLADX(i1q1, cossin1, 0);
		*src2++ = __PKHTB(q0, q1, 16);
		size -= 4;
	}
}

#define RESAMPLE_NUM_TAPS	64

// two arrays of same FIR coefficients for even and odd alignment sample
q15_t decimate_fir_coeff[2][RESAMPLE_NUM_TAPS] = {
	{ -61, -51, -69, -89, -109, -130, -148, -163, -172, -173, -165, -146, -112, -64,   0,  80, 177, 290, 417, 557, 707, 863, 1023, 1181, 1335, 1478, 1608, 1720, 1812, 1879, 1920, 1934, 1920, 1879, 1812, 1720, 1608, 1478, 1335, 1181, 1023, 863, 707, 557, 417, 290, 177,  80,   0, -64, -112, -146, -165, -173, -172, -163, -148, -130, -109, -89, -69, -51, -61, 0 },
	{ 0, -61, -51, -69, -89, -109, -130, -148, -163, -172, -173, -165, -146, -112, -64,   0,  80, 177, 290, 417, 557, 707, 863, 1023, 1181, 1335, 1478, 1608, 1720, 1812, 1879, 1920, 1934, 1920, 1879, 1812, 1720, 1608, 1478, 1335, 1181, 1023, 863, 707, 557, 417, 290, 177,  80,   0, -64, -112, -146, -165, -173, -172, -163, -148, -130, -109, -89, -69, -51, -61 }
};

am_demod_state_t am_demod_state;

volatile struct {
	uint16_t write_current;
	uint16_t write_total;
	uint16_t read_total;
	uint16_t read_current;
	uint16_t rebuffer_count;
} audio_state;

__RAMFUNC(RAM)
void resample_fir_filter()
{
	const uint32_t *coeff;
	const uint16_t *src1 = (const uint16_t *)I_FIR_STATE;
	const uint16_t *src2 = (const uint16_t *)Q_FIR_STATE;
	const uint32_t *s1, *s2;
	int32_t tail = FIR_BUFFER_SIZE;
	int32_t idx = am_demod_state.index;
	int32_t acc1, acc2;
	int i, j;
	int cur = audio_state.write_current;
	uint16_t *dest = (uint16_t *)AUDIO_BUFFER;

	while (idx < tail) {
		coeff = (uint32_t*)decimate_fir_coeff[idx % 2];
		acc1 = 0;
		acc2 = 0;
		s1 = (const uint32_t*)&src1[idx >> 1];
		s2 = (const uint32_t*)&src2[idx >> 1];
		for (j = 0; j < RESAMPLE_NUM_TAPS / 2; j++) {
			uint32_t x1 = *s1++;
			uint32_t x2 = *s2++;
			acc1 = __SMLAD(x1, *coeff, acc1);
			acc2 = __SMLAD(x2, *coeff, acc2);
			coeff++;
		}
		acc1 = __SSAT(acc1 >> 14, 16);
		acc2 = __SSAT(acc2 >> 14, 16);

#if STEREO
		dest[cur++] = acc1;
		dest[cur++] = acc2;
		audio_state.write_total += 2;
#else
		dest[cur++] = acc1;
		audio_state.write_total += 1;
#endif
		cur %= AUDIO_BUFFER_SIZE / 2;
		idx += 10; /* 1/10 decimation: 2 samples per loop */
	}

#if 0
	//am_demod_state.average = (average * 31 + (sum / count)) / 32;
	am_demod_state.average = (average * 3 + (sum / count)) / 4;
	//am_demod_state.average = (average + (sum / count)) / 2;
	if (am_demod_state.average > 3096) {
		if (fir_gain > 1)
			fir_gain -= fir_gain / 16 + 1;
		else if (fir_gain == 1 && fir_rshift < 4)
			fir_rshift++;
	} else if (average < 2048) {
		if (fir_gain < 8000)
			fir_gain += fir_gain / 16 + 1;
	} else if (fir_rshift > 0 && fir_gain > 10) {
		fir_rshift--;
	}
#endif

	audio_state.write_current = cur;
	am_demod_state.index = idx - tail;
	uint32_t *state = (uint32_t *)I_FIR_STATE;
	src1 = &src1[tail / sizeof(*src1)];
	for (i = 0; i < FIR_STATE_SIZE / sizeof(uint32_t); i++) {
		//*state++ = *src1++;
	    __asm__ volatile ("ldr r0, [%0], #+4\n" : : "r" (src1) : "r0");
	    __asm__ volatile ("str r0, [%0], #+4\n" : : "r" (state) : "r0");
	}
	state = (uint32_t *)Q_FIR_STATE;
	src2 = &src2[tail / sizeof(*src2)];
	for (i = 0; i < FIR_STATE_SIZE / sizeof(uint32_t); i++) {
		//*state++ = *src2++;
	    __asm__ volatile ("ldr r0, [%0], #+4\n" : : "r" (src2) : "r0");
	    __asm__ volatile ("str r0, [%0], #+4\n" : : "r" (state) : "r0");
	}
}

/*
void generate_test_tone(int freq)
{
	int i;
	int16_t *buf = (int16_t*)AUDIO_BUFFER;
	int samples = AUDIO_BUFFER_SIZE / 2;
	int n = freq * samples / 48000;
	for (i = 0; i < AUDIO_BUFFER_SIZE / 2; i++) {
		float res = arm_sin_f32(((float)i * 2.0 * PI * n) / 4096);
		buf[i] = (int)(res * 1500.0) + 1500;
	}
}
*/
/*
void test_cos_sin()
{
	uint16_t phase = 0;
	int i, j;
	for (j = 0; j < 16; j++) {
		for (i = 0; i < 16; i++) {
			printf("%x ", cos_sin(phase, 0));
			phase += 500;
		}
		printf("\n");
	}
}
*/

__RAMFUNC(RAM)
void DMA_IRQHandler (void)
{
  if (LPC_GPDMA->INTERRSTAT & 1)
  {
    LPC_GPDMA->INTERRCLR = 1;
  }

  if (LPC_GPDMA->INTTCSTAT & 1)
  {
	LPC_GPDMA->INTTCCLEAR = 1;

	TESTPOINT_ON();
    if ((capture_count & 1) == 0) {
    	cic_decimate(&cic_i, CAPTUREBUFFER0, CAPTUREBUFFER_SIZEHALF);
    	cic_decimate(&cic_q, CAPTUREBUFFER0, CAPTUREBUFFER_SIZEHALF);
    } else {
    	cic_decimate(&cic_i, CAPTUREBUFFER1, CAPTUREBUFFER_SIZEHALF);
    	cic_decimate(&cic_q, CAPTUREBUFFER1, CAPTUREBUFFER_SIZEHALF);
    }
#if 1
    TESTPOINT_SPIKE();
    finetune_mixer();
    TESTPOINT_SPIKE();
	resample_fir_filter();
#endif

	TESTPOINT_OFF();
    capture_count ++;

    {
    	// toggle LED with every 1024 interrupts
    	int c = capture_count % 1024;
    	if (c == 0)
    	 	LED_ON();
    	else if (c == 512)
    	 	LED_OFF();
    }
  }
#if 0
  {
	  static int first = TRUE;
	  if (first) {
		pwmout_setupdma((uint16_t*)AUDIO_BUFFER, AUDIO_BUFFER_SIZE/sizeof(uint16_t));
		first = FALSE;
	  }
  }
#endif
}

__RAMFUNC(RAM)
void I2S0_IRQHandler()
{
#if 1
	uint32_t txLevel = I2S_GetLevel(LPC_I2S0, I2S_TX_MODE);
	//TESTPOINT_ON();
	if (txLevel < 8) {
		// Fill the remaining FIFO
		int cur = audio_state.read_current;
		int16_t *buffer = (int16_t*)AUDIO_BUFFER;
		int i;
		for (i = 0; i < (8 - txLevel); i++) {
			uint32_t x = *(uint32_t *)&buffer[cur]; // read TWO samples
			LPC_I2S0->TXFIFO = x;//__PKHTB(x, x, 0);
			cur += 2;
			cur %= AUDIO_BUFFER_SIZE / 2;
			audio_state.read_total += 2;
		}
		audio_state.read_current = cur;
	}
	//TESTPOINT_OFF();
#endif
}

void
dsp_init()
{
	cic_init();
	//generate_test_tone(440);
}
